TUNE_TEMP_PATH = "tune.temp.path"
TUNE_PREFIX = "__tune_"
TUNE_DATASET_DF_PREFIX = TUNE_PREFIX + "df__"
TUNE_DATASET_PARAMS_PREFIX = TUNE_PREFIX + "params__"
TUNE_DATASET_TRIALS = TUNE_PREFIX + "trials__"

TUNE_REPORT_ID = TUNE_PREFIX + "trail_id__"
TUNE_REPORT_METRIC = TUNE_PREFIX + "metric__"
TUNE_REPORT = TUNE_PREFIX + "report__"

TUNE_REPORT_ADD_SCHEMA = (
    f"{TUNE_REPORT_ID}:str,{TUNE_REPORT_METRIC}:double,{TUNE_REPORT}:str"
)

TUNE_DATASET_DF_DEFAULT_NAME = TUNE_PREFIX + "_df_"
TUNE_DATASET_VALIDATION_DF_DEFAULT_NAME = TUNE_DATASET_DF_DEFAULT_NAME + "_validation_"

TUNE_STOPPER_DEFAULT_CHECK_INTERVAL = "60sec"

SPACE_PARAM_PREFIX = "__space_"
SPACE_MODEL_NAME = SPACE_PARAM_PREFIX + "_model"
